'''
Runs a test of the AnnotatorJ plugin's basic loading functionality. Reads a demo image.
'''

import napari
from _dock_widget import AnnotatorJ

viewer = napari.Viewer()
pluginInstance=AnnotatorJ(viewer)
pluginInstance.setTestMode(True)
viewer.window.add_dock_widget(pluginInstance,name='AnnotatorJ')
pluginInstance.openNew()
napari.run()


"""
This module is an example of a barebones numpy reader plugin for napari.

It implements the ``napari_get_reader`` hook specification, (to create
a reader plugin) but your plugin may choose to implement any of the hook
specifications offered by napari.
see: https://napari.org/docs/dev/plugins/hook_specifications.html

Replace code below accordingly.  For complete documentation see:
https://napari.org/docs/dev/plugins/for_plugin_developers.html
"""
import numpy as np
from napari_plugin_engine import napari_hook_implementation


@napari_hook_implementation
def napari_get_reader(path):
    """A basic implementation of the napari_get_reader hook specification.

    Parameters
    ----------
    path : str or list of str
        Path to file, or list of paths.

    Returns
    -------
    function or None
        If the path is a recognized format, return a function that accepts the
        same path or list of paths, and returns a list of layer data tuples.
    """
    if isinstance(path, list):
        # reader plugins may be handed single path, or a list of paths.
        # if it is a list, it is assumed to be an image stack...
        # so we are only going to look at the first file.
        path = path[0]

    # if we know we cannot read the file, we immediately return None.
    if not path.endswith(".npy"):
        return None

    # otherwise we return the *function* that can read ``path``.
    return reader_function


def reader_function(path):
    """Take a path or list of paths and return a list of LayerData tuples.

    Readers are expected to return data as a list of tuples, where each tuple
    is (data, [add_kwargs, [layer_type]]), "add_kwargs" and "layer_type" are
    both optional.

    Parameters
    ----------
    path : str or list of str
        Path to file, or list of paths.

    Returns
    -------
    layer_data : list of tuples
        A list of LayerData tuples where each tuple in the list contains
        (data, metadata, layer_type), where data is a numpy array, metadata is
        a dict of keyword arguments for the corresponding viewer.add_* method
        in napari, and layer_type is a lower-case string naming the type of layer.
        Both "meta", and "layer_type" are optional. napari will default to
        layer_type=="image" if not provided
    """
    # handle both a string and a list of strings
    paths = [path] if isinstance(path, str) else path
    # load all files into array
    arrays = [np.load(_path) for _path in paths]
    # stack arrays into single array
    data = np.squeeze(np.stack(arrays))

    # optional kwargs for the corresponding viewer.add_* method
    add_kwargs = {}

    layer_type = "image"  # optional, default is "image"
    return [(data, add_kwargs, layer_type)]



"""
This module is an example of a barebones writer plugin for napari

It implements the ``napari_get_writer`` and ``napari_write_image`` hook specifications.
see: https://napari.org/docs/dev/plugins/hook_specifications.html

Replace code below according to your needs
"""

import skimage.io
import numpy

supported_layers = ['labels']


def napari_get_writer(path, layer_types):
	# Check that only supported layers have been passed
	for x in set(layer_types):
		if x not in supported_layers:
			return None
	
	if isinstance(path, str) and path.endswith('.tiff'):
		return napari_write_labels
	else:
		return None
	return napari_write_labels


def napari_write_labels(path: str, data: numpy.ndarray, meta: dict):
	if data is None:
		return None
	out=data.astype('uint16')
	skimage.io.imsave(path,out,check_contrast=False)
	print('Saved exported image: {}'.format(path))
	print('---------------------')

	return path



import napari
from _dock_widget import AnnotatorJ
from roifile import ImagejRoi
from napari.layers import Shapes, Image, Labels, Layer

def countLayers(layers,layerType=Image):
    c=0
    for x in layers:
        if (x.__class__ is layerType):
            c+=1
    return c

viewer = napari.Viewer()
pluginInstance=AnnotatorJ(viewer)
viewer.window.add_dock_widget(pluginInstance,name='AnnotatorJ')

tmp_path="C:/work/szbk/annotatorj/testing_napari/dummying_nucleus/masks_demo/labelled_masks/"
pluginInstance.defFile='full adj.tiff'
# load dummy roi
rois=ImagejRoi.fromfile(pluginInstance.test_rois)
shapesLayer=pluginInstance.extractROIdata(rois)
n=len(shapesLayer.data)
print(f'numshapes: {n}')
# add 2nd shapes layer with these rois
pluginInstance.viewer.add_layer(shapesLayer)
prev_count_shapes=countLayers(pluginInstance.viewer.layers,layerType=Shapes)
print(f'before: {prev_count_shapes}')
#'''
pluginInstance.viewer.layers.remove('ROI')
print(f'is 2?: {countLayers(pluginInstance.viewer.layers,layerType=Shapes)}')
pluginInstance.loadRoisFromMask(tmp_path,False)
after_count_shapes=countLayers(pluginInstance.viewer.layers,layerType=Shapes)
print(f'after: {after_count_shapes}')
newShapesLayer=pluginInstance.viewer.layers[-1]
print(f'new shapes num: {len(newShapesLayer.data)}')
#'''

napari.run()


try:
    from ._version import version as __version__
except ImportError:
    __version__ = "unknown"


from ._reader import napari_get_reader
from ._writer import napari_get_writer, napari_write_labels
from ._dock_widget import AnnotatorJ,ClassesFrame,ExportFrame
#from ._function import napari_experimental_provide_function



import napari_annotatorj
import pytest
#from .test_dock_widget import im_layer_dims
from napari.layers import Shapes, Image, Labels, Layer
from napari.layers.shapes import _shapes_mouse_bindings as mouse_bindings
from roifile import ImagejRoi
import cv2
import numpy
from qtpy.QtCore import Qt,QTimer,QPoint
from qtpy.QtWidgets import QApplication
from time import sleep
import os

# helper fcns
def countLayers(layers,layerType=Image):
    c=0
    for x in layers:
        if (x.__class__ is layerType):
            c+=1
    return c

@pytest.fixture
def startAnnotatorJwidget(qtbot,make_napari_viewer):
    def doStart():
        viewer = make_napari_viewer()
        pluginInstance=napari_annotatorj.AnnotatorJ(viewer)
        qtbot.addWidget(pluginInstance)
        return pluginInstance
    return doStart

def im_layer(dims):
    # dims is like (100,100,3)
    return Image(numpy.array(numpy.random.random(dims)),name='Image')

@pytest.fixture
def write_test_file(tmp_path):
    def write(filename):
        # tmp_path is a pytest fixture, it will be cleaned up when the tests are run
        testFile=str(tmp_path/filename)
        testData=numpy.random.random((20,20))
        cv2.imwrite(testFile,testData)

        return testFile,testData

    return write

def init_annotatorj_w_image(startAnnotatorJwidget):
    pluginInstance=startAnnotatorJwidget()
    dims=(100,100,3)
    dummy_image=im_layer(dims)
    imageLayer = pluginInstance.viewer.add_layer(dummy_image)
    # mock a name so init fcn doesn't fail
    pluginInstance.destNameRaw='dummy'
    return pluginInstance

@pytest.fixture
def save_roi_2_mask(tmp_path):
    def write_mask(filename,shapesLayer,w,h):
        labels=shapesLayer.to_labels([w,h])
        outputFileName=os.path.join(tmp_path,filename)
        napari_annotatorj.ExportFrame.saveExportedImage(labels,outputFileName)
        return outputFileName,labels
    return write_mask

@pytest.fixture
def save_roi_2_text(tmp_path):
    def write_text(filename,exporter,shapesLayer,format=0,w=256,h=256):
        if format==0:
            bboxes=exporter.fillBboxList(shapesLayer,len(shapesLayer.data))
            exporter.saveExportedCSV(bboxes,os.path.join(tmp_path,filename))
            return True
        elif format==1:
            bboxes=exporter.fillBboxListYOLO(shapesLayer,len(shapesLayer.data),w,h)
            exporter.saveExportedCSVyolo(bboxes,os.path.join(tmp_path,filename))
            return True
        else:
            return False
    return write_text


def test_reset_annot_type(startAnnotatorJwidget):
    pluginInstance=startAnnotatorJwidget()
    # reset annot type
    pluginInstance.selectedAnnotationType='instance'
    pluginInstance.writeParams2File()

def test_init_fcns(startAnnotatorJwidget):
    pluginInstance=startAnnotatorJwidget()
    pluginInstance.preInitChkBxs()
    assert pluginInstance.contAssist==False, f'expected contAssist to be {False} but was {pluginInstance.contAssist}'
    assert pluginInstance.editMode==False, f'expected editMode to be {False} but was {pluginInstance.editMode}'

    pluginInstance.initChkBoxes()
    assert pluginInstance.selectedAnnotationType=='instance',f'expected selectedAnnotationType to be instance but was {pluginInstance.selectedAnnotationType}'
    assert pluginInstance.classMode==False, f'expected classMode to be {False} but was {pluginInstance.classMode}'
    assert pluginInstance.addAuto==False, f'expected addAuto to be {False} but was {pluginInstance.addAuto}'
    assert pluginInstance.chkEdit.isEnabled()==True,f'expected chkEdit enabled to be {True} but was {pluginInstance.chkEdit.isEnabled()}'
    assert pluginInstance.chckbxContourAssist.isEnabled()==True,f'expected chckbxContourAssist enabled to be {True} but was {pluginInstance.chckbxContourAssist.isEnabled()}'
    assert pluginInstance.chckbxClass.isEnabled()==True,f'expected chckbxClass enabled to be {True} but was {pluginInstance.chckbxClass.isEnabled()}'

    print('init fcns tested')


def test_opening_image_randgen(tmp_path,startAnnotatorJwidget):
    pluginInstance=startAnnotatorJwidget()
    sleep(1)
    # fake a 100x100x3 random rgb image
    dims=(100,100,3)
    expected_width=0.5
    dummy_image=im_layer(dims)
    prev_count_im=countLayers(pluginInstance.viewer.layers)
    prev_count_shapes=countLayers(pluginInstance.viewer.layers,layerType=Shapes)
    # mock openNew fcn
    imageLayer = pluginInstance.viewer.add_layer(dummy_image)
    after_count_im=countLayers(pluginInstance.viewer.layers)
    assert after_count_im==prev_count_im+1,f'expected image layer count after image addition to be {prev_count_im+1} but was {after_count_im}'
    assert prev_count_shapes==1,f'expected shapes layer count to be 1 but was {prev_count_shapes}'
    s=imageLayer.data.shape
    assert s[0]==dims[0],f'expected image to have dim1 dims[0] but was {s[0]}'
    assert s[1]==dims[1],f'expected image to have dim1 dims[1] but was {s[1]}'
    assert s[2]==dims[2],f'expected image to have dim1 dims[2] but was {s[2]}'
    pluginInstance.imgSize=s
    assert pluginInstance.imgSize==imageLayer.data.shape,f'unexpected image layer size'
    pluginInstance.defDir=tmp_path
    pluginInstance.defFile='dummy_name.png'
    pluginInstance.curPredictionImageName=pluginInstance.defFile
    pluginInstance.curPredictionImage=None
    pluginInstance.curOrigImage=None

    pluginInstance.finishOpenNewInit()
    assert pluginInstance.findROIlayer is not None,f'expected to find at least 1 ROI layer'
    # the correct annotEdgeWidth for a <300 x <300 image is 0.5
    assert pluginInstance.annotEdgeWidth==expected_width,f'expected annotEdgeWidth to be {expected_width} but was {pluginInstance.annotEdgeWidth}'
    assert pluginInstance.started==True,f'expected started to be True but was {pluginInstance.started}'


def test_opening_image_randgen_size(tmp_path,startAnnotatorJwidget):
    sizes=[(100,100,3),
        (300,300,3),(400,400,3),(500,500,3),(700,700,3),
        (1000,1000,3),(1200,1200,3),(1500,1500,3),(1800,1800,3),
        (2000,2000,3),(2300,2300,3),(3000,3000,3),(3100,3100,3)]
    expected_width=[0.5,
        0.5,1.0,1.0,1.5,
        1.5,2.0,2.0,3.0,
        3.0,5.0,5.0,7.0]
    pluginInstance=startAnnotatorJwidget()
    for s,e in zip(sizes,expected_width):
        # fake a random rgb image
        dims=s
        expected_width=e
        dummy_image=im_layer(dims)
        imageLayer = pluginInstance.viewer.add_layer(dummy_image)
        s=imageLayer.data.shape
        assert s[0]==dims[0],f'expected image to have dim1 dims[0] but was {s[0]}'
        assert s[1]==dims[1],f'expected image to have dim1 dims[1] but was {s[1]}'
        assert s[2]==dims[2],f'expected image to have dim1 dims[2] but was {s[2]}'
        pluginInstance.imgSize=s
        assert pluginInstance.imgSize==imageLayer.data.shape,f'unexpected image layer size'
        pluginInstance.defDir=tmp_path
        pluginInstance.defFile='dummy_name.png'
        pluginInstance.curPredictionImageName=pluginInstance.defFile
        pluginInstance.curPredictionImage=None
        pluginInstance.curOrigImage=None

        pluginInstance.finishOpenNewInit()
        assert pluginInstance.findROIlayer is not None,f'expected to find at least 1 ROI layer'
        assert pluginInstance.annotEdgeWidth==expected_width,f'expected annotEdgeWidth to be {expected_width} but was {pluginInstance.annotEdgeWidth}'


def test_opening_image_randgen_semantic(tmp_path,startAnnotatorJwidget):
    pluginInstance=startAnnotatorJwidget()
    sleep(1)
    pluginInstance.selectedAnnotationType='semantic'
    # fake a 100x100x3 random rgb image
    dims=(100,100,3)
    expected_width=0.5
    dummy_image=im_layer(dims)
    prev_count_im=countLayers(pluginInstance.viewer.layers)
    prev_count_shapes=countLayers(pluginInstance.viewer.layers,layerType=Shapes)
    # mock openNew fcn
    imageLayer = pluginInstance.viewer.add_layer(dummy_image)
    after_count_im=countLayers(pluginInstance.viewer.layers)
    assert after_count_im==prev_count_im+1,f'expected image layer count after image addition to be {prev_count_im+1} but was {after_count_im}'
    assert prev_count_shapes==1,f'expected shapes layer count to be 1 but was {prev_count_shapes}'
    s=imageLayer.data.shape
    pluginInstance.imgSize=s
    pluginInstance.defDir=tmp_path
    pluginInstance.defFile='dummy_name.png'
    pluginInstance.curPredictionImageName=pluginInstance.defFile
    pluginInstance.curPredictionImage=None
    pluginInstance.curOrigImage=None

    pluginInstance.finishOpenNewInit()
    #assert pluginInstance.findROIlayer is None,f'expected to find no ROI layer'
    labelLayer=pluginInstance.findLabelsLayerName(layerName='semantic')
    assert labelLayer is not None,f'expected to find semantic annotation layer'
    labelSize=labelLayer.data.shape
    assert labelSize[0]==s[0] and labelSize[1]==s[1]
    assert labelLayer.mode=='paint'
    assert labelLayer.opacity==0.5
    assert pluginInstance.started==True,f'expected started to be True but was {pluginInstance.started}'


def test_opening_image_randgen_bbox(tmp_path,startAnnotatorJwidget):
    pluginInstance=startAnnotatorJwidget()
    sleep(1)
    pluginInstance.selectedAnnotationType='bbox'
    # fake a 100x100x3 random rgb image
    dims=(100,100,3)
    expected_width=0.5
    dummy_image=im_layer(dims)
    prev_count_im=countLayers(pluginInstance.viewer.layers)
    prev_count_shapes=countLayers(pluginInstance.viewer.layers,layerType=Shapes)
    # mock openNew fcn
    imageLayer = pluginInstance.viewer.add_layer(dummy_image)
    after_count_im=countLayers(pluginInstance.viewer.layers)
    assert after_count_im==prev_count_im+1,f'expected image layer count after image addition to be {prev_count_im+1} but was {after_count_im}'
    assert prev_count_shapes==1,f'expected shapes layer count to be 1 but was {prev_count_shapes}'
    s=imageLayer.data.shape
    pluginInstance.imgSize=s
    pluginInstance.defDir=tmp_path
    pluginInstance.defFile='dummy_name.png'
    pluginInstance.curPredictionImageName=pluginInstance.defFile
    pluginInstance.curPredictionImage=None
    pluginInstance.curOrigImage=None

    pluginInstance.finishOpenNewInit()
    roiLayer=pluginInstance.findROIlayer()
    assert roiLayer is not None,f'expected to find at least 1 ROI layer'
    assert roiLayer.mode=='add_rectangle'
    assert pluginInstance.freeHandROIvis not in roiLayer.mouse_drag_callbacks
    assert pluginInstance.started==True,f'expected started to be True but was {pluginInstance.started}'

    # reset annot type
    pluginInstance.selectedAnnotationType='instance'
    pluginInstance.writeParams2File()

'''
def test_opening_image_real(tmp_path,qtbot,startAnnotatorJwidget,write_test_file):
    pluginInstance=startAnnotatorJwidget()
    file,data=write_test_file("dummy_name.png")
    pluginInstance.defDir=tmp_path
    pluginInstance.defFile='dummy_name.png'
    prev_count_im=countLayers(pluginInstance.viewer.layers)
    #qtbot.mouseClick(pluginInstance.btnOpen,Qt.LeftButton,delay=0)

    # mock clicking ok
    def handle_dialog():
        #while not pluginInstance.openQfileDialog.isVisible():
        #    pass
        while pluginInstance.openQfileDialog is None:
            pass
        #qtbot.keyPress(pluginInstance.openQfileDialog,Qt.Key_Enter)
        print(pluginInstance.openQfileDialog)
        pluginInstance.openQfileDialog.accept()
        #qtbot.keyPress(pluginInstance.openQfileDialog,Qt.Key_Enter)
        print('*/*/*/*/*/*/*/ PRESSED ENTER ON DIALOG /*/*/*/*/*/*/*/*')
    #openDialog=QApplication.focusWidget()
    #print(openDialog.__dir__())
    #openDialog.accept()
    QTimer.singleShot(500,handle_dialog)
    qtbot.mouseClick(pluginInstance.btnOpen,Qt.LeftButton,delay=0)

    after_count_im=countLayers(pluginInstance.viewer.layers)
    assert after_count_im==prev_count_im+1,f'expected image layer count after image addition to be {prev_count_im+1} but was {after_count_im}'
'''
def test_opening_test_image(qtbot,startAnnotatorJwidget):
    pluginInstance=startAnnotatorJwidget()
    sleep(10)
    pluginInstance.testMode=True
    prev_count_im=countLayers(pluginInstance.viewer.layers)
    qtbot.mouseClick(pluginInstance.btnOpen,Qt.LeftButton,delay=0)
    after_count_im=countLayers(pluginInstance.viewer.layers)
    assert after_count_im==prev_count_im+1,f'expected image layer count after image addition to be {prev_count_im+1} but was {after_count_im}'
    assert pluginInstance.curFileIdx==0
    assert pluginInstance.buttonPrev.isEnabled()==False
    assert pluginInstance.buttonNext.isEnabled()==False
    assert pluginInstance.lblCurrentFile.text()==" (1/1): img.png"
    assert pluginInstance.selectedAnnotationType=="instance"
    roiLayer=pluginInstance.findROIlayer()
    assert roiLayer.mode=="add_polygon"


def test_opening_test_image_w_ROI(qtbot,startAnnotatorJwidget):
    pluginInstance=startAnnotatorJwidget()
    sleep(10)
    pluginInstance.setTestMode(mode=True)
    qtbot.mouseClick(pluginInstance.btnOpen,Qt.LeftButton,delay=0)
    qtbot.mouseClick(pluginInstance.btnLoad,Qt.LeftButton,delay=0)
    roiLayer=pluginInstance.findROIlayer()
    assert roiLayer.mode=="add_polygon"
    assert pluginInstance.loadedROI==True
    assert pluginInstance.roiCount==49


def test_roi_loading_from_mask(tmp_path,startAnnotatorJwidget,save_roi_2_mask):
    pluginInstance=startAnnotatorJwidget()
    sleep(1)
    testName='dummy_mask.tiff'
    pluginInstance.defFile='dummy_mask.tiff'
    # load dummy roi
    rois=ImagejRoi.fromfile(pluginInstance.test_rois)
    shapesLayer=pluginInstance.extractROIdata(rois)
    assert shapesLayer is not None
    n=len(shapesLayer.data)
    assert n>0
    # add 2nd shapes layer with these rois
    pluginInstance.viewer.add_layer(shapesLayer)
    prev_count_shapes=countLayers(pluginInstance.viewer.layers,layerType=Shapes)
    assert prev_count_shapes==2
    filename,labelLayer=save_roi_2_mask(testName,shapesLayer,256,256)
    assert countLayers(pluginInstance.viewer.layers,layerType=Shapes)==2
    pluginInstance.loadRoisFromMask(tmp_path,False)
    after_count_shapes=countLayers(pluginInstance.viewer.layers,layerType=Shapes)
    assert after_count_shapes==prev_count_shapes
    newShapesLayer=pluginInstance.viewer.layers[-1]
    assert len(newShapesLayer.data)==2*n
    pluginInstance.viewer.layers.remove('ROI')
    pluginInstance.loadRoisFromMask(tmp_path,False)
    after_count_shapes=countLayers(pluginInstance.viewer.layers,layerType=Shapes)
    assert after_count_shapes==prev_count_shapes
    newShapesLayer=pluginInstance.viewer.layers[-1]
    assert len(newShapesLayer.data)==n


def test_classified_roi_loading(qtbot,startAnnotatorJwidget):
    pluginInstance=startAnnotatorJwidget()
    # use the classified roi.zip file
    pluginInstance.test_rois=pluginInstance.test_rois[:-4]+'_classes.zip'
    pluginInstance.setTestMode(True)
    qtbot.mouseClick(pluginInstance.btnLoad,Qt.LeftButton,delay=0)
    roiLayer=pluginInstance.findROIlayer()
    assert roiLayer is not None
    assert len(roiLayer.data)>0
    assert pluginInstance.startedClassifying==True


def test_add2roimanager(tmp_path,startAnnotatorJwidget):
    pluginInstance=startAnnotatorJwidget()
    sleep(1)
    # load dummy roi
    rois=ImagejRoi.fromfile(pluginInstance.test_rois)
    pluginInstance.add2RoiManager(rois)
    assert pluginInstance.manager is not None
    assert pluginInstance.roiCount>0 and pluginInstance.roiCount==len(rois)

    # init the manager too
    pluginInstance.viewer.layers.remove('ROI')
    pluginInstance.manager=None
    pluginInstance.add2RoiManager(rois)
    assert pluginInstance.manager is not None
    assert pluginInstance.roiCount>0 and pluginInstance.roiCount==len(rois)


def test_roi_loading_from_coords(tmp_path,startAnnotatorJwidget,save_roi_2_text):
    pluginInstance=startAnnotatorJwidget()
    sleep(1)
    testName='dummy_mask.csv'
    testName2='dummy_mask.txt'
    pluginInstance.defFile='dummy_mask.tiff'
    # load dummy roi
    rois=ImagejRoi.fromfile(pluginInstance.test_rois)
    shapesLayer=pluginInstance.extractROIdata(rois)
    prev_count_shapes=countLayers(pluginInstance.viewer.layers,layerType=Shapes)
    assert shapesLayer is not None
    n=len(shapesLayer.data)
    assert n>0
    assert prev_count_shapes==1
    pluginInstance.testMode=False

    # start an exporter
    pluginInstance.ExportFrame=napari_annotatorj.ExportFrame(pluginInstance.viewer,annotatorjObj=pluginInstance)
    # default COCO coords format (absolute)
    assert pluginInstance.ExportFrame.bboxFormat==0
    tmp=save_roi_2_text(testName,pluginInstance.ExportFrame,shapesLayer,0)
    assert os.path.isfile(os.path.join(tmp_path,testName))

    # see that absolute coords loading doesn't fail when no image is opened
    pluginInstance.loadRoisFromCoords(tmp_path)
    after_count_shapes=countLayers(pluginInstance.viewer.layers,layerType=Shapes)
    assert after_count_shapes==prev_count_shapes+1
    newShapesLayer=pluginInstance.viewer.layers[-1]
    assert newShapesLayer.name=='ROI'
    assert len(newShapesLayer.data)==n
    assert 'ROI' in pluginInstance.viewer.layers and 'ROI_prev' in pluginInstance.viewer.layers
    pluginInstance.viewer.layers.remove('ROI')
    assert countLayers(pluginInstance.viewer.layers,layerType=Shapes)==prev_count_shapes

    # delete the tmp file to ensure the next load doesn't attempt to load it
    os.remove(os.path.join(tmp_path,testName))
    assert not os.path.isfile(os.path.join(tmp_path,testName))

    # see that coord loading fails when no image is opened to calculate the relative coords to xy positions when saved as YOLO coords
    pluginInstance.ExportFrame.bboxFormat=1
    assert pluginInstance.ExportFrame.bboxFormat==1
    tmp=save_roi_2_text(testName2,pluginInstance.ExportFrame,shapesLayer,1,256,256) # test image size
    assert os.path.isfile(os.path.join(tmp_path,testName2))
    pluginInstance.loadRoisFromCoords(tmp_path)
    after_count_shapes=countLayers(pluginInstance.viewer.layers,layerType=Shapes)
    assert after_count_shapes==prev_count_shapes # no new loaded roi layer

    # now add the image then load the relative coords file again to succeed
    pluginInstance.testMode=True
    pluginInstance.openNew()
    # check if the image is opened
    assert countLayers(pluginInstance.viewer.layers)==1
    pluginInstance.testMode=False
    pluginInstance.defFile='dummy_mask.tiff'
    pluginInstance.loadRoisFromCoords(tmp_path)
    after_count_shapes=countLayers(pluginInstance.viewer.layers,layerType=Shapes)
    assert after_count_shapes==prev_count_shapes+2
    newShapesLayer=pluginInstance.viewer.layers[-1]
    assert len(newShapesLayer.data)==n


def test_saveData(qtbot,tmp_path,startAnnotatorJwidget,write_test_file):
    pluginInstance=init_annotatorj_w_image(startAnnotatorJwidget)
    testName='dummy.png'
    pluginInstance.defDir=tmp_path
    pluginInstance.defFile=testName
    testFile,testData=write_test_file(testName)
    assert countLayers(pluginInstance.viewer.layers)==1
    # load dummy roi
    rois=ImagejRoi.fromfile(pluginInstance.test_rois)
    shapesLayer=pluginInstance.extractROIdata(rois)
    pluginInstance.viewer.add_layer(shapesLayer)
    assert countLayers(pluginInstance.viewer.layers,layerType=Shapes)==2
    roiLayer=pluginInstance.findROIlayer()
    assert len(roiLayer.data)>0
    assert pluginInstance.stepping==False
    assert pluginInstance.startedClassifying==False
    assert pluginInstance.started==False
    assert pluginInstance.findImageLayer() is not None and pluginInstance.findImageLayer().data is not None
    assert pluginInstance.findOpenedImage()==True
    pluginInstance.started=True
    assert pluginInstance.started==True
    # save
    def handle_dialog():
        # get a reference to the dialog and handle it here
        while pluginInstance.classSelectionDialog is None:
            pass
        pluginInstance.saveClassSelectionOk()
    QTimer.singleShot(500, handle_dialog)
    qtbot.mouseClick(pluginInstance.btnSave,Qt.LeftButton,delay=1)
    assert pluginInstance.selectedClass=='normal'
    assert os.path.isfile(os.path.join(tmp_path,'normal','dummy_ROIs.zip'))
    assert pluginInstance.finishedSaving==True



def test_init_annotatorj_w_image(startAnnotatorJwidget):
    pluginInstance=startAnnotatorJwidget()
    sleep(10)
    assert pluginInstance.contAssist==False
    assert pluginInstance.classMode==False
    assert pluginInstance.editMode==False
    assert pluginInstance.chkEdit.isEnabled()==True
    assert pluginInstance.chckbxClass.isEnabled()==True
    prev_count_shapes=countLayers(pluginInstance.viewer.layers,layerType=Shapes)

    # add a dummy image layer
    # fake a 100x100x3 random rgb image
    dims=(100,100,3)
    dummy_image=im_layer(dims)
    prev_count_im=countLayers(pluginInstance.viewer.layers)
    imageLayer = pluginInstance.viewer.add_layer(dummy_image)
    after_count_im=countLayers(pluginInstance.viewer.layers)
    assert after_count_im==prev_count_im+1,f'expected image layer count after image addition to be {prev_count_im+1} but was {after_count_im}'
    return pluginInstance


# ---------------

# test checkboxes
def test_checkbox_init_contAssist(startAnnotatorJwidget):
    pluginInstance=init_annotatorj_w_image(startAnnotatorJwidget)
    pluginInstance.chckbxContourAssist.setChecked(True)
    sleep(1)
    assert pluginInstance.chckbxContourAssist.isChecked()==True
    assert pluginInstance.contAssist==True
    pluginInstance.preInitChkBxs()
    assert pluginInstance.chckbxAddAutomatically.isEnabled()==False
    assert pluginInstance.editMode==False
    assert pluginInstance.classMode==False
    assert pluginInstance.chckbxClass.isEnabled()==False


def test_checkbox_init_editMode(startAnnotatorJwidget):
    pluginInstance=init_annotatorj_w_image(startAnnotatorJwidget)
    pluginInstance.chkEdit.setChecked(True)
    sleep(1)
    assert pluginInstance.chkEdit.isChecked()==True
    assert pluginInstance.editMode==True
    pluginInstance.preInitChkBxs()
    assert pluginInstance.chckbxAddAutomatically.isEnabled()==False
    assert pluginInstance.contAssist==False
    assert pluginInstance.classMode==False
    assert pluginInstance.chckbxClass.isEnabled()==False
    assert pluginInstance.chckbxContourAssist.isEnabled()==False


def test_checkbox_init_classMode(startAnnotatorJwidget):
    pluginInstance=init_annotatorj_w_image(startAnnotatorJwidget)
    pluginInstance.chckbxClass.setChecked(True)
    sleep(1)
    assert pluginInstance.chckbxClass.isChecked()==True
    assert pluginInstance.classMode==True
    assert pluginInstance.editMode==False
    pluginInstance.initChkBoxes()
    assert pluginInstance.chkEdit.isEnabled()==False
    assert pluginInstance.editMode==False
    assert pluginInstance.contAssist==False
    assert pluginInstance.classMode==True
    assert pluginInstance.chckbxContourAssist.isEnabled()==False


def test_checkbox_init_editMode2(startAnnotatorJwidget):
    pluginInstance=init_annotatorj_w_image(startAnnotatorJwidget)
    pluginInstance.chkEdit.setChecked(True)
    sleep(1)
    assert pluginInstance.chkEdit.isChecked()==True
    assert pluginInstance.classMode==False
    assert pluginInstance.editMode==True
    pluginInstance.initChkBoxes()
    assert pluginInstance.chckbxClass.isEnabled()==False
    assert pluginInstance.editMode==True
    assert pluginInstance.contAssist==False
    assert pluginInstance.classMode==False
    assert pluginInstance.chckbxContourAssist.isEnabled()==False


def test_checkbox_init_constAssist2(startAnnotatorJwidget):
    pluginInstance=init_annotatorj_w_image(startAnnotatorJwidget)
    pluginInstance.chckbxContourAssist.setChecked(True)
    sleep(1)
    assert pluginInstance.chckbxContourAssist.isChecked()==True
    assert pluginInstance.classMode==False
    assert pluginInstance.editMode==False
    assert pluginInstance.contAssist==True
    pluginInstance.initChkBoxes()
    assert pluginInstance.chkEdit.isEnabled()==False
    assert pluginInstance.editMode==False
    assert pluginInstance.contAssist==True
    assert pluginInstance.classMode==False
    assert pluginInstance.chckbxClass.isEnabled()==False


def test_checkbox_init_semantic(startAnnotatorJwidget):
    pluginInstance=init_annotatorj_w_image(startAnnotatorJwidget)
    pluginInstance.selectedAnnotationType='semantic'
    sleep(1)
    assert pluginInstance.classMode==False
    assert pluginInstance.editMode==False
    assert pluginInstance.contAssist==False
    pluginInstance.initChkBoxes()
    assert pluginInstance.chkEdit.isEnabled()==False
    assert pluginInstance.editMode==False
    assert pluginInstance.contAssist==False
    assert pluginInstance.classMode==False
    assert pluginInstance.chckbxContourAssist.isEnabled()==False
    assert pluginInstance.chckbxClass.isEnabled()==False


def test_checkbox_init_bbox(startAnnotatorJwidget):
    pluginInstance=init_annotatorj_w_image(startAnnotatorJwidget)
    pluginInstance.selectedAnnotationType='bbox'
    sleep(1)
    assert pluginInstance.classMode==False
    assert pluginInstance.editMode==False
    assert pluginInstance.contAssist==False
    pluginInstance.initChkBoxes()
    assert pluginInstance.chkEdit.isEnabled()==False
    assert pluginInstance.editMode==False
    assert pluginInstance.contAssist==False
    assert pluginInstance.classMode==False
    assert pluginInstance.chckbxContourAssist.isEnabled()==False
    assert pluginInstance.chckbxClass.isEnabled()==True


def test_checkbox_contAssist(startAnnotatorJwidget):
    pluginInstance=init_annotatorj_w_image(startAnnotatorJwidget)
    assert countLayers(pluginInstance.viewer.layers)==1
    prev_count_shapes=countLayers(pluginInstance.viewer.layers,layerType=Shapes)

    # start testing the checkbox
    pluginInstance.chckbxContourAssist.setChecked(True)
    sleep(1)
    assert pluginInstance.chckbxContourAssist.isChecked()==True
    assert pluginInstance.contAssist==True
    assert pluginInstance.classMode==False
    assert pluginInstance.editMode==False
    assert pluginInstance.chkEdit.isEnabled()==False
    assert pluginInstance.chckbxClass.isEnabled()==False
    after_count_shapes=countLayers(pluginInstance.viewer.layers,layerType=Shapes)
    assert after_count_shapes==prev_count_shapes+1,f'expected shapes layer count after contour assist init to be {prev_count_shapes+1} but was {after_count_shapes}'
    contAssistLayer=pluginInstance.viewer.layers[-1]
    assert contAssistLayer.name=='contourAssist'
    assert contAssistLayer.mode=='add_polygon'

    # check default unchecked states
    pluginInstance.chckbxContourAssist.setChecked(False)
    sleep(1)
    assert countLayers(pluginInstance.viewer.layers,layerType=Shapes)==prev_count_shapes
    assert pluginInstance.chckbxContourAssist.isChecked()==False
    assert pluginInstance.contAssist==False

    # remove image to trigger fail
    pluginInstance.viewer.layers.remove('Image')
    assert countLayers(pluginInstance.viewer.layers)==0
    pluginInstance.chckbxContourAssist.setChecked(True)
    sleep(1)
    assert pluginInstance.chckbxContourAssist.isChecked()==True
    assert pluginInstance.contAssist==False


def test_checkbox_classMode(startAnnotatorJwidget):
    pluginInstance=init_annotatorj_w_image(startAnnotatorJwidget)
    assert countLayers(pluginInstance.viewer.layers)==1
    prev_count_shapes=countLayers(pluginInstance.viewer.layers,layerType=Shapes)
    num_dw = len(pluginInstance.viewer.window._dock_widgets)

    # start testing the checkbox
    pluginInstance.chckbxClass.setChecked(True)
    sleep(1)
    assert pluginInstance.chckbxClass.isChecked()==True
    assert pluginInstance.contAssist==False
    assert pluginInstance.classMode==True
    assert pluginInstance.editMode==False
    assert pluginInstance.chkEdit.isEnabled()==True
    assert pluginInstance.chckbxContourAssist.isEnabled()==False
    after_count_shapes=countLayers(pluginInstance.viewer.layers,layerType=Shapes)
    assert after_count_shapes==prev_count_shapes,f'expected shapes layer count after contour assist init to be {prev_count_shapes} but was {after_count_shapes}'
    roiLayer=pluginInstance.findROIlayer()
    assert roiLayer.name=='ROI'
    assert roiLayer.mode=='select'

    assert len(pluginInstance.viewer.window._dock_widgets)==num_dw+1
    assert pluginInstance.classesFrame is not None,f'Failed to open Classes widget'
    assert len(pluginInstance.classFrameNames)==2

    # check default unchecked states
    pluginInstance.chckbxClass.setChecked(False)
    sleep(1)
    assert pluginInstance.chckbxClass.isChecked()==False
    assert pluginInstance.classMode==False
    assert pluginInstance.chckbxContourAssist.isEnabled()==True
    roiLayer=pluginInstance.findROIlayer()
    assert roiLayer.mode=='add_polygon'
    assert mouse_bindings.select not in roiLayer.mouse_drag_callbacks
    assert pluginInstance.customShapesLayerSelect not in roiLayer.mouse_drag_callbacks

    # check semantic disabled state
    pluginInstance.selectedAnnotationType='semantic'
    pluginInstance.chckbxClass.setChecked(True)
    sleep(1)
    assert pluginInstance.chckbxClass.isChecked()==True
    assert pluginInstance.classMode==False

    # reset
    pluginInstance.selectedAnnotationType='instance'
    pluginInstance.chckbxClass.setChecked(False)

    # remove image to trigger fail
    pluginInstance.viewer.layers.remove('Image')
    assert countLayers(pluginInstance.viewer.layers)==0
    pluginInstance.chckbxClass.setChecked(True)
    sleep(1)
    assert pluginInstance.chckbxClass.isChecked()==True
    assert pluginInstance.classMode==False



def test_the_fcns(startAnnotatorJwidget):
    #pluginInstance=startAnnotatorJwidget()
    #test_init_fcns(startAnnotatorJwidget)
    #test_opening_image_randgen(startAnnotatorJwidget,im_layer)
    pass


'''
Runs the AnnotatorJ plugin.
'''

import napari
from _dock_widget import AnnotatorJ

viewer = napari.Viewer()
pluginInstance=AnnotatorJ(viewer)
viewer.window.add_dock_widget(pluginInstance,name='AnnotatorJ')
napari.run()


import napari_annotatorj
import pytest
from time import sleep
from napari_annotatorj._dock_widget import OptionsFrame,ColourSelector,TrainWidget,HelpWidget,Q3DWidget,FileListWidget

# this is your plugin name declared in your napari.plugins entry point
MY_PLUGIN_NAME = "napari-annotatorj"
# the name of your widget(s)
MY_WIDGET_NAMES = ["AnnotatorJ","AnnotatorJExport"]

'''
@pytest.mark.parametrize("widget_name", MY_WIDGET_NAMES)
def test_something_with_viewer(widget_name, make_napari_viewer, napari_plugin_manager):
    napari_plugin_manager.register(napari_annotatorj, name=MY_PLUGIN_NAME)
    viewer = make_napari_viewer()
    num_dw = len(viewer.window._dock_widgets)
    viewer.window.add_plugin_dock_widget(
        plugin_name=MY_PLUGIN_NAME, widget_name=widget_name
    )
    assert len(viewer.window._dock_widgets) == num_dw + 1
'''

#'''
@pytest.fixture
def test_annotatorj_widget():
    def test_main_widget_loading(qtbot,make_napari_viewer):
        viewer = make_napari_viewer()
        pluginInstance=napari_annotatorj.AnnotatorJ(viewer)
        qtbot.addWidget(pluginInstance)
        num_dw = len(viewer.window._dock_widgets)
        viewer.window.add_dock_widget(pluginInstance,name='AnnotatorJ')
        assert len(viewer.window._dock_widgets) == num_dw + 1

        # give some time for the plugin init function to complete successfully on a thread loading the unet model
        sleep(10)
        #assert pluginInstance.trainedUNetModel is not None, f'trainedUnetModel is None'
        return pluginInstance

    return test_main_widget_loading
#'''

@pytest.mark.parametrize("widget_name", MY_WIDGET_NAMES)
def test_all_widget_loading(widget_name, make_napari_viewer, napari_plugin_manager):
    napari_plugin_manager.register(napari_annotatorj, name=MY_PLUGIN_NAME)
    viewer = make_napari_viewer()
    num_dw = len(viewer.window._dock_widgets)
    viewer.window.add_plugin_dock_widget(
        plugin_name=MY_PLUGIN_NAME, widget_name=widget_name
    )
    assert len(viewer.window._dock_widgets) == num_dw + 1

def test_widget_loading_extras(make_napari_viewer,qtbot):
    viewer=make_napari_viewer()
    pluginInstance=napari_annotatorj.AnnotatorJ(viewer)
    qtbot.addWidget(pluginInstance)
    num_dw = len(viewer.window._dock_widgets)
    viewer.window.add_dock_widget(pluginInstance,name='AnnotatorJ')
    assert len(viewer.window._dock_widgets) == num_dw + 1

    # add classes widget
    pluginInstance.classesFrame=napari_annotatorj.ClassesFrame(pluginInstance.viewer,pluginInstance)
    assert len(viewer.window._dock_widgets) == num_dw + 2

    # add options widget
    pluginInstance.openOptionsFrame()
    assert len(viewer.window._dock_widgets) == num_dw + 3

    # add colours widget
    pluginInstance.addColourWidget()
    assert len(viewer.window._dock_widgets) == num_dw + 4

    # add training widget
    pluginInstance.openTrainWidget()
    assert len(viewer.window._dock_widgets) == num_dw + 5

    # add help widget
    pluginInstance.openHelpWidgetDock()
    assert len(viewer.window._dock_widgets) == num_dw + 6

    # add 3d widget
    pluginInstance.open3DWidget()
    assert len(viewer.window._dock_widgets) == num_dw + 7

    # add filelist widget
    pluginInstance.openFileListWidget()
    assert len(viewer.window._dock_widgets) == num_dw + 8


@pytest.fixture
def im_layer():
    return Image(np.array(np.random.random((100,100,3))),name='Image')


"""
This module is an example of a barebones QWidget plugin for napari

It implements the ``napari_experimental_provide_dock_widget`` hook specification.
see: https://napari.org/docs/dev/plugins/hook_specifications.html

Replace code below according to your needs.
"""
from napari_plugin_engine import napari_hook_implementation
from qtpy.QtWidgets import QWidget, QHBoxLayout, QPushButton
from magicgui import magic_factory


class ExampleQWidget(QWidget):
    # your QWidget.__init__ can optionally request the napari viewer instance
    # in one of two ways:
    # 1. use a parameter called `napari_viewer`, as done here
    # 2. use a type annotation of 'napari.viewer.Viewer' for any parameter
    def __init__(self, napari_viewer):
        super().__init__()
        self.viewer = napari_viewer

        btn = QPushButton("Click me!")
        btn.clicked.connect(self._on_click)

        self.setLayout(QHBoxLayout())
        self.layout().addWidget(btn)

    def _on_click(self):
        print("napari has", len(self.viewer.layers), "layers")


@magic_factory
def example_magic_widget(img_layer: "napari.layers.Image"):
    print(f"you have selected {img_layer}")


@napari_hook_implementation
def napari_experimental_provide_dock_widget():
    # you can return either a single widget, or a sequence of widgets
    return [ExampleQWidget, example_magic_widget]



from model import *
from data import *

#os.environ["CUDA_VISIBLE_DEVICES"] = "0"


data_gen_args = dict(rotation_range=0.2,
                    width_shift_range=0.05,
                    height_shift_range=0.05,
                    shear_range=0.05,
                    zoom_range=0.05,
                    horizontal_flip=True,
                    fill_mode='nearest')
myGene = trainGenerator(2,'data/membrane/train','image','label',data_gen_args,save_to_dir = None)

model = unet()
model_checkpoint = ModelCheckpoint('unet_membrane.hdf5', monitor='loss',verbose=1, save_best_only=True)
model.fit_generator(myGene,steps_per_epoch=300,epochs=1,callbacks=[model_checkpoint])

testGene = testGenerator("data/membrane/test")
results = model.predict_generator(testGene,30,verbose=1)
saveResult("data/membrane/test",results)


import napari
from _dock_widget import AnnotatorJ,ExportFrame
from roifile import ImagejRoi
from napari.layers import Shapes, Image, Labels, Layer
import os,sys

def countLayers(layers,layerType=Image):
    c=0
    for x in layers:
        if (x.__class__ is layerType):
            c+=1
    return c

viewer = napari.Viewer()
pluginInstance=AnnotatorJ(viewer)
viewer.window.add_dock_widget(pluginInstance,name='AnnotatorJ')

tmp_path="C:/work/szbk/annotatorj/testing_napari/dummying_nucleus/coording/"
pluginInstance.defFile='dummy_mask.tiff'
filename='dummy_mask.csv'
filename2='dummy_mask.txt'
# load dummy roi
rois=ImagejRoi.fromfile(pluginInstance.test_rois)
shapesLayer=pluginInstance.extractROIdata(rois)
n=len(shapesLayer.data)
print(f'numshapes: {n}')
prev_count_shapes=countLayers(pluginInstance.viewer.layers,layerType=Shapes)
print(f'before: {prev_count_shapes}')

if os.path.isfile(os.path.join(tmp_path,filename)):
    os.remove(os.path.join(tmp_path,filename))
if os.path.isfile(os.path.join(tmp_path,filename2)):
    os.remove(os.path.join(tmp_path,filename2))

pluginInstance.ExportFrame=ExportFrame(pluginInstance.viewer,annotatorjObj=pluginInstance)
bboxes=pluginInstance.ExportFrame.fillBboxList(shapesLayer,len(shapesLayer.data))
pluginInstance.ExportFrame.saveExportedCSV(bboxes,os.path.join(tmp_path,filename))
success=os.path.isfile(os.path.join(tmp_path,filename))
print(f'wrote file: {success}')

pluginInstance.loadRoisFromCoords(tmp_path)
print(f'loaded: {countLayers(pluginInstance.viewer.layers,layerType=Shapes)}')
#'''
after_count_shapes=countLayers(pluginInstance.viewer.layers,layerType=Shapes)
print(f'after: {after_count_shapes}')
newShapesLayer=pluginInstance.viewer.layers[-1]
print(f'new shapes num: {len(newShapesLayer.data)}')
#'''

# remove it
print(f'layers before REMOVE: {pluginInstance.viewer.layers}')
pluginInstance.viewer.layers.remove('ROI')
print(f'--------------- REMOVING ROI LAYER -----------------')
print(f'after remove layer count: {countLayers(pluginInstance.viewer.layers,layerType=Shapes)}')
print(f'layers after REMOVE: {pluginInstance.viewer.layers}')
os.remove(os.path.join(tmp_path,filename))

# write yolo file
bboxes=pluginInstance.ExportFrame.fillBboxListYOLO(shapesLayer,n,256,256)
pluginInstance.ExportFrame.saveExportedCSVyolo(bboxes,os.path.join(tmp_path,filename2))
success=os.path.isfile(os.path.join(tmp_path,filename2))
print(f'wrote second file: {success}')

pluginInstance.loadRoisFromCoords(tmp_path)
print(f'loaded 2nd: {countLayers(pluginInstance.viewer.layers,layerType=Shapes)}')
newShapesLayer=pluginInstance.viewer.layers[-1]
print(f'numshapes 2nd: {len(newShapesLayer.data)}')

pluginInstance.testMode=True
pluginInstance.openNew()
print(f'opened image')
pluginInstance.testMode=False
pluginInstance.defFile='dummy_mask.tiff'
pluginInstance.loadRoisFromCoords(tmp_path)
print(f'loaded 3rd: {countLayers(pluginInstance.viewer.layers,layerType=Shapes)}')
newShapesLayer=pluginInstance.viewer.layers[-1]
print(f'numshapes 3rd: {len(newShapesLayer.data)}')


napari.run()







from __future__ import print_function
from keras.preprocessing.image import ImageDataGenerator
import numpy as np 
import os
import glob
import skimage.io as io
import skimage.transform as trans

Sky = [128,128,128]
Building = [128,0,0]
Pole = [192,192,128]
Road = [128,64,128]
Pavement = [60,40,222]
Tree = [128,128,0]
SignSymbol = [192,128,128]
Fence = [64,64,128]
Car = [64,0,128]
Pedestrian = [64,64,0]
Bicyclist = [0,128,192]
Unlabelled = [0,0,0]

COLOR_DICT = np.array([Sky, Building, Pole, Road, Pavement,
                          Tree, SignSymbol, Fence, Car, Pedestrian, Bicyclist, Unlabelled])


def adjustData(img,mask,flag_multi_class,num_class):
    if(flag_multi_class):
        img = img / 255
        mask = mask[:,:,:,0] if(len(mask.shape) == 4) else mask[:,:,0]
        new_mask = np.zeros(mask.shape + (num_class,))
        for i in range(num_class):
            #for one pixel in the image, find the class in mask and convert it into one-hot vector
            #index = np.where(mask == i)
            #index_mask = (index[0],index[1],index[2],np.zeros(len(index[0]),dtype = np.int64) + i) if (len(mask.shape) == 4) else (index[0],index[1],np.zeros(len(index[0]),dtype = np.int64) + i)
            #new_mask[index_mask] = 1
            new_mask[mask == i,i] = 1
        new_mask = np.reshape(new_mask,(new_mask.shape[0],new_mask.shape[1]*new_mask.shape[2],new_mask.shape[3])) if flag_multi_class else np.reshape(new_mask,(new_mask.shape[0]*new_mask.shape[1],new_mask.shape[2]))
        mask = new_mask
    elif(np.max(img) > 1):
        img = img / 255
        mask = mask /255
        mask[mask > 0.5] = 1
        mask[mask <= 0.5] = 0
    return (img,mask)



def trainGenerator(batch_size,train_path,image_folder,mask_folder,aug_dict,image_color_mode = "grayscale",
                    mask_color_mode = "grayscale",image_save_prefix  = "image",mask_save_prefix  = "mask",
                    flag_multi_class = False,num_class = 2,save_to_dir = None,target_size = (256,256),seed = 1):
    '''
    can generate image and mask at the same time
    use the same seed for image_datagen and mask_datagen to ensure the transformation for image and mask is the same
    if you want to visualize the results of generator, set save_to_dir = "your path"
    '''
    image_datagen = ImageDataGenerator(**aug_dict)
    mask_datagen = ImageDataGenerator(**aug_dict)
    image_generator = image_datagen.flow_from_directory(
        train_path,
        classes = [image_folder],
        class_mode = None,
        color_mode = image_color_mode,
        target_size = target_size,
        batch_size = batch_size,
        save_to_dir = save_to_dir,
        save_prefix  = image_save_prefix,
        seed = seed)
    mask_generator = mask_datagen.flow_from_directory(
        train_path,
        classes = [mask_folder],
        class_mode = None,
        color_mode = mask_color_mode,
        target_size = target_size,
        batch_size = batch_size,
        save_to_dir = save_to_dir,
        save_prefix  = mask_save_prefix,
        seed = seed)
    train_generator = zip(image_generator, mask_generator)
    for (img,mask) in train_generator:
        img,mask = adjustData(img,mask,flag_multi_class,num_class)
        yield (img,mask)



def testGenerator(test_path,num_image = 30,target_size = (256,256),flag_multi_class = False,as_gray = True):
    for i in range(num_image):
        img = io.imread(os.path.join(test_path,"%d.png"%i),as_gray = as_gray)
        img = img / 255
        img = trans.resize(img,target_size)
        img = np.reshape(img,img.shape+(1,)) if (not flag_multi_class) else img
        img = np.reshape(img,(1,)+img.shape)
        yield img


def geneTrainNpy(image_path,mask_path,flag_multi_class = False,num_class = 2,image_prefix = "image",mask_prefix = "mask",image_as_gray = True,mask_as_gray = True):
    image_name_arr = glob.glob(os.path.join(image_path,"%s*.png"%image_prefix))
    image_arr = []
    mask_arr = []
    for index,item in enumerate(image_name_arr):
        img = io.imread(item,as_gray = image_as_gray)
        img = np.reshape(img,img.shape + (1,)) if image_as_gray else img
        mask = io.imread(item.replace(image_path,mask_path).replace(image_prefix,mask_prefix),as_gray = mask_as_gray)
        mask = np.reshape(mask,mask.shape + (1,)) if mask_as_gray else mask
        img,mask = adjustData(img,mask,flag_multi_class,num_class)
        image_arr.append(img)
        mask_arr.append(mask)
    image_arr = np.array(image_arr)
    mask_arr = np.array(mask_arr)
    return image_arr,mask_arr


def labelVisualize(num_class,color_dict,img):
    img = img[:,:,0] if len(img.shape) == 3 else img
    img_out = np.zeros(img.shape + (3,))
    for i in range(num_class):
        img_out[img == i,:] = color_dict[i]
    return img_out / 255



def saveResult(save_path,npyfile,flag_multi_class = False,num_class = 2):
    for i,item in enumerate(npyfile):
        img = labelVisualize(num_class,COLOR_DICT,item) if flag_multi_class else item[:,:,0]
        io.imsave(os.path.join(save_path,"%d_predict.png"%i),img)


from PyInstaller.utils.hooks import collect_data_files
from vispy.app.backends import CORE_BACKENDS

datas = collect_data_files('vispy')

hiddenimports = ["vispy.ext._bundled.six"]

# adding all backends is required for vispy.sys_info() to work
hiddenimports += ["vispy.app.backends." + b[1] for b in CORE_BACKENDS]
hiddenimports += ["vispy.app.backends._test"]



# help the manual startup script import:
try:
	from .unet.model import *
except ImportError as e:
	try:
		from unet.model import *
	except Exception as e:
		print(e)
		exit()

import skimage.io as io
import cv2
import numpy as np

def importArgs():
	import argparse
	parser = argparse.ArgumentParser()
	parser.add_argument('--results', type=str, required=True)
	parser.add_argument('--train', type=str, required=False)
	parser.add_argument('--val', type=str, default=None)
	parser.add_argument('--test', type=str, default=None)
	parser.add_argument('--epochs', type=int, default=5)
	parser.add_argument('--steps', type=int, default=300)
	parser.add_argument('--model', type=str, default=None)
	parser.add_argument('--batch', type=int, default=2)
	parser.add_argument('--gpu', type=str)
	parser.add_argument('--write', type=bool, default=True)
	parser.add_argument('--size', type=int, default=256)
	args = parser.parse_args()

	return args


def loadUnetModel(modelName,importMode=0):
	model=None
	# load json and create model
	print('starting loadUnetModel...')
	try:
		if importMode==0:
			# import from '.json' + '_weights.h5' files
			json_file = open(modelName+'.json', 'r')
			loaded_model_json = json_file.read()
			json_file.close()
			model = model_from_json(loaded_model_json)
			# load weights into new model
			model.load_weights(modelName+'_weights.h5')
		elif importMode==1:
			# import from a single '.hdf5' file
			model=load_model(modelName+'.hdf5')
	except Exception as e:
		print(f'Could not load model {modelName}_weights.h5')
		print(e)
		raise(e)
	else:
		print(f"Loaded model from disk: {modelName}_weights.h5")
	finally:
		return model


def loadUnetModelSetGpu(modelName,importMode=0,gpuSetting='0'):
	setGpu(gpuSetting)
	model=None
	# load json and create model
	print('starting loadUnetModel...')
	try:
		if importMode==0:
			# import from '.json' + '_weights.h5' files
			json_file = open(modelName+'.json', 'r')
			loaded_model_json = json_file.read()
			json_file.close()
			model = model_from_json(loaded_model_json)
			# load weights into new model
			model.load_weights(modelName+'_weights.h5')
		elif importMode==1:
			# import from a single '.hdf5' file
			model=load_model(modelName+'.hdf5')
	except Exception as e:
		print(f'Could not load model {modelName}_weights.h5')
		print(e)
		raise(e)
	else:
		print(f"Loaded model from disk: {modelName}_weights.h5")
	finally:
		return model
		

def trainIfNoModel(args):
	# if model exists skip training and only predict test images
	if os.path.isfile(args.model+'.json') and os.path.isfile(args.model+'_weights.h5'):
		# load json and create model
		model=loadUnetModel(args.model)

	else:
		# do training first
		# help the manual startup script import:
		try:
			from .unet.data import trainGenerator
		except ImportError as e:
			try:
				from unet.data import trainGenerator
			except Exception as e:
				print(e)
				exit()
		
		data_gen_args = dict(rotation_range=0.2,
							width_shift_range=0.05,
							height_shift_range=0.05,
							shear_range=0.05,
							zoom_range=0.05,
							horizontal_flip=True,
							fill_mode='nearest')
		myGene = trainGenerator(args.batch,args.train,'images','unet_masks',data_gen_args,save_to_dir = None,target_size = (256,256),image_color_mode = "rgb")

		model = unet(input_size = (256,256,3))
		model_checkpoint = ModelCheckpoint(args.model+'.hdf5', monitor='loss',verbose=1, save_best_only=True)
		model.fit_generator(myGene,steps_per_epoch=args.steps,epochs=args.epochs,callbacks=[model_checkpoint])

		# save model as json
		model_json = model.to_json()
		with open(args.model+'.json', 'w') as f:
			f.write(model_json)

		# save weights too
		model.save_weights(args.model+'_weights.h5')

	return model


def predictUnet(model,imgPath,target_size=(256,256),flag_multi_class=False):
	if isinstance(imgPath,str):
		img = io.imread(imgPath,as_gray = False) #as_gray = True #as_gray = as_gray
	elif isinstance(imgPath,np.ndarray):
		# already an image matrix
		img=imgPath
		pass
	else:
		# this should never happen
		print('Input is neither a path nor an image')
		return None

	'''
	if len(img.shape)==3:
		# still RGB, convert
		img=skimage.color.rgb2gray(img)
	'''
	if len(img.shape)==3:
		img=img[:,:,0:3]
	elif len(img.shape)==2:
		tmp=np.zeros((img.shape[0],img.shape[1],3),img.dtype)
		for k in range(2):
			tmp[:,:,k]=img
		img=tmp

	print(img.shape)
	img = img / 255
	orig_size=img.shape
	img = trans.resize(img,target_size)
	img = np.reshape(img,img.shape+(1,)) if ((not flag_multi_class) and len(img.shape)==2) else img
	img = np.reshape(img,(1,)+img.shape)

	# predict
	pred_image=model.predict(img,batch_size=1,verbose=1) # batch_size=1
	pred_image2write=pred_image[0,:,:,0]
	pred_image2write=cv2.resize(pred_image2write,(orig_size[1],orig_size[0]),interpolation = cv2.INTER_CUBIC)

	return pred_image2write


def predictNSaveImage(model,test_path,out_path,target_size = (256,256),flag_multi_class = False,num_class = 2,write_images=True):
	imageFiles = [f for f in os.listdir(test_path) if os.path.isfile(os.path.join(test_path,f))]
	imcount = len(imageFiles)
	print('Found ',imcount,' files in ',test_path)

	os.makedirs(name=out_path, exist_ok=True)

	if not write_images:
		predictions=[]

	for index, imageFile in enumerate(imageFiles):
		# read image
		print("Image:", str(index + 1), "/", str(imcount), "(", imageFile, ")")
		
		pred_image2write=predictUnet(model,os.path.join(test_path,imageFile),target_size)

		if write_images:
			# write output image
			io.imsave(os.path.join(out_path,imageFile),pred_image2write)
		else:
			predictions.append(pred_image2write)

	if not write_images:
		return predictions
	else:
		return None


def predictUnetCustomSize(model,imgPath,target_size=(256,256),flag_multi_class=False):
	if isinstance(imgPath,str):
		img = io.imread(imgPath,as_gray = False) #as_gray = True #as_gray = as_gray
	elif isinstance(imgPath,np.ndarray):
		# already an image matrix
		img=imgPath
		pass
	else:
		# this should never happen
		print('Input is neither a path nor an image')
		return None

	if len(img.shape)==3:
		img=img[:,:,0:3]
	elif len(img.shape)==2:
		tmp=np.zeros((img.shape[0],img.shape[1],3),img.dtype)
		for k in range(2):
			tmp[:,:,k]=img
		img=tmp

	print(img.shape)
	img = img / 255
	orig_size=img.shape

	s=orig_size[0:2]
	if s[0]>256 or s[1]>256:
		if s[0]>=512 or s[1]>=512:
			if s[0]>=1024 or s[1]>=1024:
				if s[0]<2048 or s[1]<2048:
					img=trans.resize(img,(int(s[0]/2),int(s[1]/2)))
					s=img.shape
					s=s[0:2]
				else:
					# just resize to 1024x1024
					maxdim=1024
					img=trans.resize(img,(maxdim,maxdim))
					s=img.shape
					s=s[0:2]
			# crop it
			maxdim = min(s)
			temp = maxdim / 2 ** 8
			if temp != int(temp):
				maxdim = (int(temp) + 1) * 2 ** 8
				temp=int(temp) + 1
			img=trans.resize(img,(maxdim,maxdim))
			pred_image2write=np.zeros((maxdim,maxdim),dtype='float32')
			crops=[]
			for i in range(int(temp)):
				for j in range(int(temp)):
					cropped=img[(i)*256:(i+1)*256,(j)*256:(j+1)*256,:]
					#crops.append(cropped)
					pred_cropped=calcPred(model,cropped,flag_multi_class)
					# stitch
					pred_image2write[(i)*256:(i+1)*256,(j)*256:(j+1)*256]=pred_cropped
			'''
			pred_cropped=calcPredCrops(model,crops,flag_multi_class)
			for i in range(int(temp)):
				for j in range(int(temp)):
					# stitch
					pred_image2write[(i)*256:(i+1)*256,(j)*256:(j+1)*256]=pred_cropped[(i)*temp+j]
			'''
		else:
			# just resize to 256x256
			maxdim=256
			img=trans.resize(img,(maxdim,maxdim))

			pred_image2write=calcPred(model,img,flag_multi_class)
	else:
		# small image, just resize (upscale) to 256x256
		img = trans.resize(img,(256,256))
		pred_image2write=calcPred(model,img,flag_multi_class)

	#print('pred done')
	pred_image2write=cv2.resize(pred_image2write,(orig_size[1],orig_size[0]),interpolation = cv2.INTER_CUBIC)
	#pred_image2write=predictUnet(model,img,target_size)

	return pred_image2write


def calcPred(model,img,flag_multi_class):
	img = np.reshape(img,img.shape+(1,)) if ((not flag_multi_class) and len(img.shape)==2) else img
	img = np.reshape(img,(1,)+img.shape)

	# predict
	pred_image=model.predict(img,batch_size=1,verbose=0) # batch_size=1,verbose=1
	pred_image2write=pred_image[0,:,:,0]
	return pred_image2write


def calcPredCrops(model,img,flag_multi_class):
	if isinstance(img,list) and len(img)>1:
		for idx,val in enumerate(img):
			tmp=img[idx]
			tmp = np.reshape(tmp,tmp.shape+(1,)) if ((not flag_multi_class) and len(tmp.shape)==2) else tmp
			tmp = np.reshape(tmp,(1,)+tmp.shape)
			img[idx]=tmp
		bsize=len(img)
		preds=[]
		for el in img:
			pred_image=model.predict(el,batch_size=1,verbose=1)
			preds.append(pred_image[0,:,:,0])
		return preds
	else:
		img = np.reshape(img,img.shape+(1,)) if ((not flag_multi_class) and len(img.shape)==2) else img
		img = np.reshape(img,(1,)+img.shape)
		bsize=1

	# predict
	pred_image=model.predict(img,batch_size=bsize,verbose=1) # batch_size=1
	pred_image2write=pred_image[0,:,:,0]
	return pred_image2write


def predictNSaveImageCrop(model,test_path,out_path,target_size = (256,256),flag_multi_class = False,num_class = 2,write_images=True):
	imageFiles = [f for f in os.listdir(test_path) if os.path.isfile(os.path.join(test_path,f))]
	imcount = len(imageFiles)
	print('Found ',imcount,' files in ',test_path)

	os.makedirs(name=out_path, exist_ok=True)

	if not write_images:
		predictions=[]

	for index, imageFile in enumerate(imageFiles):
		# read image
		print("Image:", str(index + 1), "/", str(imcount), "(", imageFile, ")")
		
		pred_image2write=predictUnetCustomSize(model,os.path.join(test_path,imageFile),target_size=target_size,flag_multi_class=flag_multi_class)

		if write_images:
			# write output image
			io.imsave(os.path.join(out_path,imageFile),pred_image2write)
		else:
			predictions.append(pred_image2write)

	if not write_images:
		return predictions
	else:
		return None


def predict(model,args):
	# start prediction either after training or model loading
	preds=predictNSaveImage(model,args.test,args.results,write_images=args.write) #,target_size=(args.size,args.size))


def predictCustomSize(args):
	# start prediction either after training or model loading
	model=unet(pretrained_weights=None,input_size=(args.size,args.size,3))
	print(f'size: {args.size}')
	model.load_weights(args.model+'_weights.h5')
	preds=predictNSaveImage(model,args.test,args.results,write_images=args.write,target_size=(args.size,args.size))


def predictCustomSizeCrop(model,args):
	# start prediction either after training or model loading
	preds=predictNSaveImageCrop(model,args.test,args.results,write_images=args.write,target_size=(args.size,args.size))


def setGpu(gpuSetting='0'):
	if gpuSetting is not None:
		import tensorflow as tf
		if gpuSetting=='cpu':
			tf.config.set_visible_devices([], 'GPU')
			print('Using CPU for U-Net predictions')
		else:
			# check if there is a gpu at all
			g=tf.config.list_physical_devices('GPU')
			if len(g)>int(gpuSetting):
				tf.config.set_visible_devices(g[int(gpuSetting)], 'GPU')
				print(f'Using GPU {gpuSetting} for U-Net predictions')
			else:
				tf.config.set_visible_devices([], 'GPU')
				print(f'The selected GPU device ({gpuSetting}) is not available, using CPU for U-Net predictions')


def callPredictUnet(modelName,imageName,gpuSetting='0'):
	model=loadUnetModel(modelName)
	setGpu(gpuSetting)
	pred=predictUnet(model,imageName)
	return pred


def callPredictUnetLoaded(model,imageName,gpuSetting='0'):
	setGpu(gpuSetting)
	pred=predictUnet(model,imageName)
	return pred


def callPredictUnetLoadedNoset(model,imageName,target_size=(256,256)):
	pred=predictUnet(model,imageName)
	return pred


def callPredictUnetLoadedNosetCustomSize(model,imageName):
	pred=predictUnetCustomSize(model,imageName)
	return pred


def estimateImgSize(self,img):
	resSize=(256,256)
	if img is not None:
		s=img.shape
		maxdim = max(s)
		temp = maxdim / 2 ** 6
		if temp != int(temp):
			maxdim = (int(temp) + 1) * 2 ** 6
		resSize=(maxdim,maxdim)
	else:
		#debug:
		print(f'img is None')
	print(f'set size: ({resSize})')
	return resSize


def main():
	args=importArgs()
	
	setGpu(args.gpu)
	
	if args.size==256:
		model=trainIfNoModel(args)
		preds=predict(model,args)
	else:
		#preds=predictCustomSize(args)
		model=trainIfNoModel(args)
		preds=predictCustomSizeCrop(model,args)
	print('Finished predicting images')

	return preds


if __name__ == '__main__':
	main()